# -*- coding: utf-8 -*-
"""openepda.updk.py

This file contains tools to work with uPDK files.

Author: Dima Pustakhod
Copyright: 2020, TU/e - PITC and authors
"""
from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from functools import lru_cache
from io import StringIO
from numbers import Real
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from .expr import TNestedRealList
from .expr import evaluate_expression
from .imports import HAS_NAZCA_DESIGN
from .imports import dump_long
from .imports import dump_short
from .imports import safe_load
from .validators import _check_building_block
from .validators import _check_parameter
from .validators import _check_pin
from .validators import is_updk_sbb_valid
from .geometry import isclose


if HAS_NAZCA_DESIGN:
    from .imports import nd

DEFAULT_BB_OUTLINE_LAYER = 1020
DEFAULT_BB_INFO_LAYER = 1025
DEFAULT_BB_PARAMS_LAYER = 1026
DEFAULT_PIN_LAYER = 1004
DEFAULT_PIN_INFO_LAYER = 1006
DEFAULT_BBOX_LAYER = 1021


def _check_or_add_xsection(xs_name, layer=DEFAULT_PIN_LAYER):
    if not HAS_NAZCA_DESIGN:
        print("This function requires nazca package.")
        return None

    try:
        nd.get_xsection(xs_name)
    except Exception:
        nd.add_layer(name=xs_name, layer=layer, accuracy=0.05)
        nd.add_xsection(name=xs_name)
        nd.add_layer2xsection(xsection=xs_name, layer=xs_name, accuracy=0.05)


def _get_or_add_layer(layer_name, layer=DEFAULT_BB_OUTLINE_LAYER):
    if not HAS_NAZCA_DESIGN:
        print("This function requires nazca package.")
        return None

    # Instead of try-except block I want to check if the layer is present,
    # see the commented if-else below.
    # Can not be done now because nazca's `get_layers()` does not work in
    # v0.5.10. `overwrite=True` is used to suppress the Error messages.
    try:
        res = nd.add_layer(
            name=layer_name, layer=layer, accuracy=0.05, overwrite=True
        )
    except Exception:
        res = nd.get_layer(layer_name)

    # if layer_name in nd.get_layers():
    #     res = layer_name
    # else:
    #     res = nd.add_layer(name=layer_name, layer=layer, accuracy=0.05)

    assert res == layer_name
    return res


class ParametrizedEntity(ABC):
    """Interface for classes with parameters.
    """

    @abstractmethod
    def get_parameter_values(self) -> Optional[Dict[str, Real]]:
        raise NotImplementedError


class ParametrizedChildEntity(ParametrizedEntity):
    """Mixin class implementing access to parent parameters

    """

    def __init__(self, parent: ParametrizedEntity = None):
        self._parent = parent

    def get_parameter_values(self):
        """Implementation of the interface method, using parent parameters.

        Returns
        -------

        """
        if not hasattr(self, "_parent"):
            self.__init__()
        elif self._parent:
            return self._parent.get_parameter_values()
        else:
            return {}


def ensure_num(
    func: Callable[[ParametrizedEntity], Any]
) -> Callable[[ParametrizedEntity], TNestedRealList]:
    """Decorator to ensure numeric type of the returned values.

    This decorator is used to ensure that the returned value has a numeric
    type. It passes the returned value through the evaluate_expression()
    function.

    This decorator is to be used on methods only, as it required the first
    argument of the decorated function to accept the class instance.

    If your would like to wrap a property, first apply this decorator, and
    then apply @property decorator:

    ```python
    class A():

        @property
        @ensure_num
        def value(self):
            return 100
    ```

    Parameters
    ----------
    func : Callable[[ParametrizedEntity], Any]
        method to be wrapped

    Returns
    -------
    func : Callable[[ParametrizedEntity], TNestedRealList]
        wrapped method
    """

    def wrapper(inst):
        res = func(inst)
        if isinstance(res, Real):
            # This check is also performed in evaluate_expression(),
            # but we keep it here for no reason.
            pass
        else:
            params = inst.get_parameter_values()
            res = evaluate_expression(expr=res, params=params)
        return res

    return wrapper


class BBox(ParametrizedChildEntity):
    """Class to handle building block bounding boxes.

    Bounding box is a rectangle aligned with XY axis, and is characterized
    by its south-west and north-east corners.
    """

    def __init__(
        self,
        sw: Tuple[Real, Real],
        ne: Tuple[Real, Real],
        parent: Optional["BB"] = None,
    ):
        super().__init__(parent)
        assert sw[0] < ne[0]
        assert sw[1] < ne[1]
        self._sw = sw
        self._ne = ne

    @property
    def length(self) -> Real:
        """Bounding box size along X-axis

        Returns
        -------
        Real
        """
        length = self._ne[0] - self._sw[0]
        assert length >= 0
        return length

    @property
    def width(self) -> Real:
        """Bounding box size along Y-axis

        Returns
        -------
        Real
        """
        w = self._ne[1] - self._sw[1]
        assert w >= 0
        return w

    @property
    def sw(self) -> Tuple[Real, Real]:
        return self._sw

    @property
    def ne(self) -> Tuple[Real, Real]:
        return self._ne

    @property
    def nw(self) -> Tuple[Real, Real]:
        return self._sw[0], self._ne[1]

    @property
    def se(self) -> Tuple[Real, Real]:
        return self._ne[0], self._sw[1]

    @staticmethod
    def from_points(
        pts: Iterable[Tuple[Real, Real]], parent: ParametrizedEntity = None
    ):
        """Create a bounding box from a list of points.

        Points can describe any polygon.

        Parameters
        ----------
        pts: Iterable[Tuple[Real, Real]]
            List of (x, y) coordinates of the points.
        parent : Optional[ParametrizedEntity]
            Parent for the bounding box.

        Returns
        -------
        BBox
            Minimal bounding box which includes all points.
        """
        params = parent.get_parameter_values() if parent else {}
        x_coords = [evaluate_expression(p[0], params=params) for p in pts]
        y_coords = [evaluate_expression(p[1], params=params) for p in pts]

        x_min = min(x_coords)
        x_max = max(x_coords)
        y_min = min(y_coords)
        y_max = max(y_coords)
        return BBox(sw=(x_min, y_min), ne=(x_max, y_max), parent=parent)

    @property
    def center(self) -> Tuple[Real, Real]:
        return (
            (self._sw[0] + self._ne[0]) // 2,
            (self._sw[1] + self._ne[1]) // 2,
        )

    def __eq__(self, other: "BBox") -> bool:
        sw = isclose(self._sw, other._sw, abs_tol=1e-10, rel_tol=1e-10)
        ne = isclose(self._ne, other._ne, abs_tol=1e-10, rel_tol=1e-10)
        return sw and ne

    def __str__(self):
        return f'<BBox (sw={self.sw}, ne={self.ne})>'


class Pin(ParametrizedChildEntity):
    DIR_SYMBOLS = defaultdict(
        lambda: "?",
        {"0.0": "->", "90.0": "/\\", "180.0": "<-", "270.0": r"\/"},
    )

    def __init__(
        self,
        name: str,
        pin_data: Dict,
        skip_checks: bool = False,
        parent: Optional["BB"] = None,
    ):
        super().__init__(parent)
        self.name = name
        self._is_valid = skip_checks

        self._data = None
        self.data = pin_data

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, pin_data):
        if self._data:
            raise UserWarning(
                "Can not change the existing Pin. Create a new one instead."
            )
        self._data = pin_data

        if not self._is_valid:
            try:
                if self._parent:
                    bb_name = self._parent.name
                    bb_params = list(self._parent.get_parameter_values().keys())
                else:
                    bb_name = "NOT DEFINED"
                    bb_params = []

                _check_pin(self.name, self._data, bb_name, bb_params)
            except Exception as e:
                self._data = None
                raise ValueError("Error validating the pin data: {}.".format(e))

    @property
    def xsection(self) -> str:
        return self._data["xsection"]

    @property
    @ensure_num
    def width(self) -> Real:
        width = self._data["width"]
        return width

    @property
    def doc(self) -> str:
        return self._data["doc"]

    def __str__(self):
        a_str = "{:.1f}".format(float(self.xya[2] % 360))
        direction = self.DIR_SYMBOLS[a_str]

        return "Pin {} {} (w={}, xs={})".format(
            self.name, direction, self.width, self.xsection
        )

    @property
    @ensure_num
    def xya(self) -> Tuple[Real, Real, Real]:
        return tuple(self._data["xya"])

    @property
    def info(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "xsection": self.xsection,
            "width": self.width,
            "xya": self.xya,
        }


class Parameter(object):
    def __init__(
        self,
        name: str,
        parameter_data: Dict,
        skip_checks: bool = False,
        parent: Optional["BB"] = None,
    ):
        self._parent = parent
        self.name = name
        self._is_valid = skip_checks

        self._data = None
        self.data = parameter_data

        self._value = None
        self.value = self.default_value

    @property
    def typ(self) -> str:
        t = self.data["type"]
        assert t in ("float", "str", "bool", "int")
        return t

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, parameter_data):
        if self._data:
            raise UserWarning(
                "Can not change the existing Parameter. "
                "Create a new one instead."
            )
        self._data = parameter_data

        if not self._is_valid:
            try:
                if self._parent:
                    bb_name = self._parent.name
                else:
                    bb_name = "NOT DEFINED"
                _check_parameter(self.name, self._data, bb_name)
            except Exception as e:
                self._data = None
                raise ValueError(
                    "Error validating the parameter data: {}.".format(e)
                )

    @property
    def default_value(self) -> Real:
        return self._data["value"]

    @property
    def min_value(self) -> Optional[Real]:
        try:
            return self._data["min"]
        except KeyError:
            raise AttributeError(
                f"Parameter '{self.name}' is of '{self.typ}' type and "
                f"it does not have min value."
            )

    @property
    def max_value(self) -> Optional[Real]:
        try:
            return self._data["max"]
        except KeyError:
            raise AttributeError(
                f"Parameter '{self.name}' is of '{self.typ}' type and "
                f"it does not have max value."
            )

    @property
    def value(self) -> Real:
        if not self._value:
            self.value = self.default_value
        return self._value

    @value.setter
    def value(self, value):
        if self.typ in ("int", "float"):
            if self.min_value > value:
                raise ValueError(
                    f"Value {value} for parameter {self.name} is "
                    f"too small. Minimum value is {self.min_value}."
                )
            if value > self.max_value:
                raise ValueError(
                    f"Value {value} for parameter {self.name} is "
                    f"too large. Maximum value is {self.max_value}."
                )
            self._value = value
        elif self.typ in ("bool",):
            if not isinstance(value, bool):
                raise ValueError(
                    f"Value for parameter {self.name} should be boolean. "
                    f"{value} (type: {type(value)}) is given."
                )
            self._value = value
        elif self.typ in ("str",):
            if not isinstance(value, str):
                raise ValueError(
                    f"Value for parameter {self.name} should be string. "
                    f"{value} (type: {type(value)}) is given."
                )
            self._value = value


class BB(ParametrizedEntity):
    """Building block class to handle the uPDK block data.

    """

    CELL_TYPES = {0: "static", 1: "p-cell"}

    def __init__(self, name: str, bb_data: dict, skip_checks: bool = False):
        """Initialize new building block.

        Parameters
        ----------
        name : str
            name of the building block
        bb_data : dict
            building block data. A complex nested structure defined by
            the uPDK format.
        skip_checks : bool
            set to True if you have performed the validation and sure
            that the data is correct, expressions can be evaluated, etc.
        """
        self.name = name
        self._is_valid = skip_checks

        self._data = None
        self.data = bb_data
        self._cells: Dict[str, "nd.Cell"] = {}  # keeps {name: Cell} pairs

        # {pin_name: (pin_data, Pin)}: this structure is used to store
        # pin data and Pin instances together in a single dictionary. Pin
        # instances are not created before they are accessed by get_pin()
        # method.
        self._pins: Dict[str, Tuple[Dict, Optional[Pin]]] = {
            k: (v, None) for k, v in self.data["pins"].items()
        }

        # {parameter_name: (parameter_data, Parameter)}: This structure is
        # used to store parameter data and Parameter instances together in a
        # single dictionary. Parameter instances are not created before they are
        # accessed by get_parameter() method.
        if self.is_pcell:
            self._params: Dict[str, Tuple[Dict, Optional[Parameter]]] = {
                k: (v, None) for k, v in self.data["parameters"].items()
            }
        else:
            self._params = {}

    @property
    def full_name(self):
        """Return full name unique for a given set of parameters.

        Returns
        -------
        full_name : str
            full BB name. Takes into account current parameter values.
        """
        if self.is_pcell:
            values = self.get_parameter_values()
            if values:
                suffix = ", ".join(
                    "{}={}".format(p_name, v) for p_name, v in values.items()
                )
                full = f"{self.name}: {suffix}"
            else:
                full = self.name
        else:
            full = self.name
        return full

    @property
    def data(self) -> Dict:
        return self._data

    @data.setter
    def data(self, bb_data):
        if self._data:
            raise UserWarning(
                "Can not change the existing BB. Create a new one instead."
            )
        self._data = bb_data

        if not self._is_valid:
            try:
                _check_building_block(self.name, self._data)
            except Exception as e:
                self._data = None
                raise ValueError("Error validating the BB data: {}.".format(e))

    @property
    def is_pcell(self) -> bool:
        """Flag if the building block is parametric.

        Returns
        -------
        bool
            True if the building block is a p-cell, False if it is static.
        """
        return True if self._data["parameters"] else False

    @property
    def cell_type(self) -> str:
        return BB.CELL_TYPES[self.is_pcell]

    @property
    def current_cell(self) -> Optional["nd.Cell"]:
        return self.get_cell(self.full_name)

    @property
    def cell_names(self) -> Tuple[str]:
        """Return names of the existing cells
        """
        return tuple(sorted(self._cells.keys()))

    def get_cell(self, name) -> Optional["nd.Cell"]:
        """Get cell by the name.

        For p-cells, it provides access to previously generated cells with
        other parameters.

        Parameters
        ----------
        name : str
            The name of the generated cell.

        Returns
        -------
        cell : Optional[nazca.Cell]
        """
        if name in self.cell_names:
            return self._cells[name]
        else:
            return None

    @property
    def n_pins(self) -> int:
        return len(self._pins)

    @property
    @lru_cache(maxsize=1)
    def pin_names(self) -> Tuple[str]:
        return tuple(sorted(self._pins.keys()))

    def get_pin_data(self, name) -> Dict:
        if name not in self.pin_names:
            raise ValueError(
                f"Pin '{name}' is not defined in the {self.name} BB."
            )

        return self._pins[name][0]

    def get_pin(self, name) -> "Pin":
        """Get pin by name

        Parameters
        ----------
        name : str
            pin name

        Returns
        -------
        Pin : Pin
            Pin instance
        """
        if name not in self.pin_names:
            raise ValueError(
                f"Pin '{name}' is not defined in the {self.name} BB."
            )

        if self._pins[name][1]:
            pin = self._pins[name][1]
        else:
            data = self.get_pin_data(name)
            pin = Pin(name, data, parent=self)
            self._pins.update({name: (data, pin)})
        return pin

    @property
    @lru_cache(maxsize=1)
    def parameter_names(self) -> Union[Tuple[str], Tuple]:
        if self.is_pcell:
            return tuple(sorted(self._params.keys()))
        else:
            return ()

    def get_parameter_data(self, name) -> Dict:
        if name in self.parameter_names:
            return self._params[name][0]
        else:
            raise ValueError(
                f"Parameter '{name}' is not defined in the {self.name} BB."
            )

    def get_parameter(self, name) -> "Parameter":
        """Get parameter by name

        Parameters
        ----------
        name : str
            parameter name

        Returns
        -------
        parameter : Parameter
            parameter instance
        """
        if name not in self.parameter_names:
            raise ValueError(
                f"Parameter '{name}' is not defined in the {self.name} BB."
            )
        if self._params[name][1]:
            parameter = self._params[name][1]
        else:
            data = self.get_parameter_data(name)
            parameter = Parameter(name, data)
            self._params.update({name: (data, parameter)})
        return parameter

    def get_parameter_value(self, name) -> Real:
        return self.get_parameter(name).value

    def get_parameter_values(self) -> Dict[str, Real]:
        values = {
            name: self.get_parameter_value(name)
            for name in self.parameter_names
        }

        return values

    def __str__(self) -> str:
        return "<BB {} ({}, {} pins)>".format(
            self.name, self.cell_type, self.n_pins
        )

    @property
    @ensure_num
    def bb_outline(self) -> List[Tuple[Real, Real]]:
        return [(x, y) for x, y in self.data["bbox"]]

    @property
    def bbox(self) -> "BBox":
        return BBox.from_points(self.bb_outline, parent=self)

    def get_or_make_cell(self) -> Optional["nd.Cell"]:
        """Create a cell for the building block.

        If the cell already exists, it will be returned.
        This method runs only if nazca package is installed.
        The cell is also added to private attribute _cells.

        Returns
        -------
        cell : Optional[nazca.Cell]
            None is returned in case if nazca is not installed,
            of if error occurs.
        """
        if not HAS_NAZCA_DESIGN:
            print("This function requires nazca package.")
            return None
        # if self.is_pcell:
        #     print(f"Cannot create a cell for a p-cell ({self.name}).")
        #     return None
        cell = self.get_cell(self.full_name)
        if not cell:
            cell = self._make_cell()
        return cell

    def _make_cell(self, add_foundry_params=False):
        """Make a new cell.

        This is not a safe method: if the cell already exists, a new one with
        a different name will be created. Use get_or_make_cell() instead.
        """
        if not HAS_NAZCA_DESIGN:
            print("This function requires nazca package.")
            return None

        try:
            # Create cross-sections
            for pin_name in self.pin_names:
                pin = self.get_pin(pin_name)
                _check_or_add_xsection(pin.xsection)

            with nd.Cell(self.full_name, hashme=False) as C:
                # C.default_pins(bb_data["pin_in"], bb_data["pin_out"])
                # C.autobbox = True
                # C.store_pins = True
                # C.version = version

                # Pins
                pin_names = []
                layer = _get_or_add_layer(
                    "pin_info_layer", DEFAULT_PIN_INFO_LAYER
                )
                for pin_name in self.pin_names:
                    pin = self.get_pin(pin_name)
                    nd.Pin(
                        name=str(pin),
                        xs=pin.xsection,
                        width=float(pin.width),
                        remark=pin.doc,
                    ).put(*pin.xya)

                    pin_info_str = StringIO()
                    dump_short(pin.info, pin_info_str)
                    pin_info_str = pin_info_str.getvalue()
                    nd.Annotation(layer=layer, text=pin_info_str).put(*pin.xya)

                    pin_names.append(str(pin))
                nd.put_stub(pin_names)

                # BB outline
                layer = _get_or_add_layer("bb_outline_layer")
                nd.Polygon(points=self.bb_outline, layer=layer).put(0)

                # BB bounding box
                bbox = self.bbox
                nd.put_boundingbox(
                    "org",
                    length=float(bbox.length),
                    width=float(bbox.width),
                    raise_pins=False,
                    move=(bbox.sw[0], bbox.sw[1], 0),
                    align="lb",
                )

                # Parameter values
                pv = self.get_parameter_values()

                if pv:
                    pv_str = StringIO()
                    dump_long(pv, pv_str)
                    pv_str = pv_str.getvalue()
                    layer = _get_or_add_layer(
                        "bb_params_layer", DEFAULT_BB_PARAMS_LAYER
                    )
                    nd.Annotation(layer=layer, text=pv_str).put(
                        bbox.center[0], bbox.center[1], 0
                    )
                if add_foundry_params:
                    C.parameters = pv
                    nd.put_parameters(
                        pv, pin=(bbox.center[0], bbox.center[1], 0)
                    )

                # Cell info
                info = self.get_info()
                info_str = StringIO()
                dump_long(info, info_str)
                info_str = info_str.getvalue()
                layer = _get_or_add_layer(
                    "bb_info_layer", DEFAULT_BB_INFO_LAYER
                )
                nd.Annotation(layer=layer, text=info_str).put(0)

            # C._add_bbox()
            # if bb_text[bn]:
            #     nd.text(
            #         bb_text[bn], height=3, align="cc", layer="DeepLogic"
            #     ).put(C.pin["bc"].move(-3, 0, 90))

            self._cells.update({C.cell_name: C})
            return C
        except Exception as e:
            print(f"Error creating cell for BB '{self.full_name}': {e}.")
            return None

    @property
    def version(self) -> Optional[str]:
        try:
            return self._data["version"]
        except KeyError:
            return None

    def get_info(self):
        info = {
            "cell_name": self.name,
            "pdk_version": None,
            "bb_version": self.version,
            "bb_owner": None,
        }
        return info

    def set_parameters(self, **kwargs):
        for param_name, value in kwargs.items():
            self.set_parameter(param_name, value)

    def set_predefined_parameters(self, which="default"):
        for param_name in self.parameter_names:
            self.set_predefined_parameter(param_name, which=which)

    def set_predefined_parameter(self, name, which="default"):
        parameter = self.get_parameter(name)

        if which == "default":
            value = parameter.default_value
        elif which == "min":
            try:
                value = parameter.min_value
            except KeyError:
                value = parameter.default_value
        elif which == "max":
            try:
                value = parameter.max_value
            except KeyError:
                value = parameter.default_value
        else:
            raise ValueError(f"Unknown option for which: '{which}'.")

        self.set_parameter(name, value)

    def set_parameter(self, name, value):
        self.get_parameter(name).value = value
        # Reset caches here if needed
        # bbox, outline, pins


class UPDK(object):
    """Class to handle the uPDK file data.
    """

    def __init__(self, updk_data):
        self._data = None
        self.data = updk_data
        self._cells: Dict[str, "nd.Cell"] = {}

        # {bb_name: (bb_data, BB)}: This structure is used to store BB data and
        # BB instances together in a single dictionary. BB instances are not
        # created before they are accessed by get_building_block() method.
        self._bbs: Dict[str, Tuple[Dict, Optional[BB]]] = {
            k: (v, None) for k, v in self.data["blocks"].items()
        }

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, updk_data):
        is_updk_sbb_valid(updk_data, raise_error=True)
        self._data = updk_data

    @staticmethod
    def from_file(filename):
        """Create a UPDK instance from a uPDK file.

        Parameters
        ----------
        filename
            str or a path-like object

        Returns
        -------
        updk : UPDK
        """
        with open(filename) as s:
            updk_data = safe_load(s)
        return UPDK(updk_data)

    @property
    @lru_cache(maxsize=1)
    def building_block_names(self) -> Tuple[str]:
        return tuple(sorted(self._bbs.keys()))

    def get_building_block_data(self, name) -> Dict:
        if name not in self.building_block_names:
            raise ValueError(
                f"Building block '{name}' is not defined in the uPDK."
            )

        return self._bbs[name][0]

    def get_building_block(self, name) -> BB:
        """Get building block by name

        Parameters
        ----------
        name : str
            building block name

        Returns
        -------
        bb : BB
            building block instance
        """
        if name not in self.building_block_names:
            raise ValueError(
                f"Building block '{name}' is not defined in the uPDK."
            )

        if self._bbs[name][1]:
            bb = self._bbs[name][1]
        else:
            data = self.get_building_block_data(name)
            bb = BB(name, data)
            self._bbs.update({name: (data, bb)})
        return bb

    def get_cells(self, bb_name) -> List["nd.Cell"]:
        """All generated cells for a given building block.
        """
        bb = self.get_building_block(bb_name)
        return [bb.get_cell(cell_name) for cell_name in bb.cell_names]

    @property
    def cells(self) -> List["nd.Cell"]:
        """All generated cells for all building blocks.
        """
        cells = []
        for bb_name in self.building_block_names:
            cells += self.get_cells(bb_name)

        return cells

    def make_cell(self, bb_name) -> Optional[str]:
        """Make a cell for a bb with currently set parameters.
        """
        bb = self.get_building_block(bb_name)
        return bb.get_or_make_cell()

    def make_cells(self) -> bool:
        """Create cells for the building blocks.

        This method runs only if nazca package is installed.
        Only static cells are processed now. The created cells are
        available via UPDK.cells property.

        Returns
        -------
        result : bool
            True if the cells were created successfully
        """
        if not HAS_NAZCA_DESIGN:
            print("This function requires nazca package.")
            return False

        for bb_name in self.building_block_names:
            self.make_cell(bb_name)

        return True

    def export_cells(self, filename) -> bool:
        """Export building block cells to a GDSII file.

        This method runs only if nazca package is installed.
        The cells should be first created using UPDK.make_cells() method.

        Parameters
        ----------
        filename : str
            The target GDSII filename.

        Returns
        -------
        result : bool
            True if the cells were created successfully
        """
        if not HAS_NAZCA_DESIGN:
            print("This function requires nazca package.")
            return False

        if not self.cells:
            print(
                "No cells were created for this PDK. "
                "Run UPDK.make_sells() first."
            )
            return False

        nd.export_gds(self.cells, filename=filename)
        return True



# -*- coding: utf-8 -*-
"""openepda.main.py

This module contains utility functions and openEPDA data file writer and
loader.

Author: Dima Pustakhod
Copyright: 2020, TU/e - PITC and authors
"""
import collections.abc as coll
import os
from csv import QUOTE_NONNUMERIC
from datetime import datetime

import numpy as np
import pandas as pd
from pandas.errors import ParserError
from ruamel.yaml import YAML


def is_list_or_tuple(x):
    """ Check if a variable is either a tuple or a list

    Returns
    -------
    bool
        True if x is a tuple or list, or a subclass of these; False otherwise

    Examples
    --------
    >>> i = 1; f = 1.1; st = 's'; bs = b'S'
    >>> d = {1: 1, 2: 2}; s = {1, 2, 3}
    >>> l = [1, 2]; t = (1, 2)
    >>> arr = np.asarray([1, 2, 3]); m = np.array([1, 2, 3])
    >>> is_list_or_tuple(i), is_list_or_tuple(f)
    (False, False)
    >>> is_list_or_tuple(st), is_list_or_tuple(bs)
    (False, False)
    >>> is_list_or_tuple(d), is_list_or_tuple(s)
    (False, False)
    >>> is_list_or_tuple(l), is_list_or_tuple(t)
    (True, True)
    >>> is_list_or_tuple(arr), is_list_or_tuple(m)
    (False, False)
    """

    is_sequence = isinstance(x, coll.Sequence)
    is_str = isinstance(x, (str, coll.ByteString))
    return is_sequence and not is_str


def is_array_like(x):
    """ Check if a variable is array-like

    Returns
    -------
    bool
        True if x is a tuple, a list, a numpy.ndarray, or a subclass of these;
        False otherwise

    Examples
    --------
    >>> i = 1; f = 1.1; st = 's'; bs = b'S'
    >>> d = {1: 1, 2: 2}; s = {1, 2, 3}
    >>> l = [1, 2]; t = (1, 2)
    >>> arr = np.asarray([1, 2, 3]); m = np.array([1, 2, 3])
    >>> is_array_like(i), is_array_like(f)
    (False, False)
    >>> is_array_like(st), is_array_like(bs)
    (False, False)
    >>> is_array_like(d), is_array_like(s)
    (False, False)
    >>> is_array_like(l), is_array_like(t)
    (True, True)
    >>> is_array_like(arr), is_array_like(m)
    (True, True)
    """
    x_is_list_or_tuple = is_list_or_tuple(x)
    is_ndarray = isinstance(x, np.ndarray)
    return x_is_list_or_tuple or is_ndarray


class OpenEpdaDataLoader(object):
    def __init__(self):
        super().__init__()

    def read_file(self, fname):

        with open(fname, "r") as f:

            lines = []
            separator_index = 0
            for i, line in enumerate(f):
                if line != "---\n":
                    lines.append(line)
                else:
                    separator_index = i
                    break

        if separator_index == 0:
            raise ValueError(
                'File "{}" does not confirm to the OpenEpda '
                'Data format: "---" line is missing. File will '
                "be skipped.".format(os.path.basename(fname))
            )

        yaml = YAML(typ="safe")
        yaml.default_flow_style = False
        yaml.explicit_end = None

        data = "\n".join(lines)
        file_data = yaml.load(data)

        try:
            data = pd.read_csv(fname, header=separator_index + 1)

            for col_name in data.columns:
                file_data.update({col_name: data[col_name].values})
        except ParserError:
            pass

        return file_data


def split_array_data(data, how="max_length"):
    """

    Parameters
    ----------
    how : str
        'max_length|most_common'

    Examples
    --------
    >>> split_array_data({2: [1, 2], '2a': [1, 4]})
    ({}, {2: [1, 2], '2a': [1, 4]})

    >>> split_array_data({2: [1, 2], 3: [1, 2, 3], '3a': [1, 2, 4]})
    ({2: [1, 2]}, {3: [1, 2, 3], '3a': [1, 2, 4]})

    >>> split_array_data({2: [1, 2, 3], 3: [1, 2], '3a': [1, 2]}, how='most_common')
    ({2: [1, 2, 3]}, {3: [1, 2], '3a': [1, 2]})
    """
    if len(data) == 0:
        return data, {}

    lengths = list(map(len, data.values()))

    if how == "max_length":
        l_csv = max(lengths)
    elif how == "most_common":
        l_csv = max(set(lengths), key=lengths.count)
    else:
        raise ValueError("Unknown value for how: {}".format(how))

    csv_data = {}
    meta_data = {}

    for k, v in data.items():
        if len(v) == l_csv:
            csv_data.update({k: v})
        else:
            meta_data.update({k: v})

    return meta_data, csv_data


class OpenEpdaDataDumper(object):
    def __init__(self, csv_data="max_length", float_format=None):
        super().__init__()
        self._csv_data = csv_data
        self._float_format = float_format

    def write(self, stream, **data):
        """
        Parameters
        ----------
        stream : Any

        data : dict

        """
        yaml_data = {"_timestamp": datetime.now().isoformat()}
        array_data = {}

        for k, v in data.items():
            if is_array_like(v):
                try:
                    l = v.tolist()
                except:
                    l = list(v)
                array_data.update({k: l})
            else:
                yaml_data.update({k: v})

        meta_data, csv_data_dict = split_array_data(
            array_data, how=self._csv_data
        )
        yaml_data.update(meta_data)

        # csv_data = np.array(list(csv_data_dict.values())).T
        # csv_data_keys = list(csv_data_dict.keys())

        # self._l.debug('yaml keys: {}'.format(yaml_data.keys()))
        # self._l.debug('yaml type: {}'.format(type(yaml_data)))
        # self._l.debug('csv keys: {}, shape: {}'.format(csv_data_keys,
        #                                                csv_data.shape))

        stream.write("# OpenEPDA DATA FORMAT v.0.1\n")

        yaml = YAML()
        yaml.default_flow_style = False
        yaml.explicit_end = None
        yaml.dump(yaml_data, stream)

        stream.write("---\n")

        df_csv = pd.DataFrame(csv_data_dict)
        # np.savetxt(stream, csv_data, delimiter=',',
        #            header=','.join(['"{}"'.format(k) for k in csv_data_keys]),
        #            comments='', fmt=self._float_format)
        df_csv.to_csv(
            stream,
            sep=",",
            quoting=QUOTE_NONNUMERIC,
            decimal=".",
            index=False,
            float_format=self._float_format,
        )



# -*- coding: utf-8 -*-
"""openepda.expr.py

This file contains a parser of analytic expressions.

The parser can be used to parse the analytic expressions and thus
check their correctness in a safe way, and also to evaluate the
expressions.

The grammar is intended to be very similar to the tinyexpr library
grammar (https://github.com/codeplea/tinyexpr).

Author: Dima Pustakhod
Copyright: 2020, TU/e - PITC and authors
"""
import re
from collections import Counter
from functools import partial
from importlib import import_module
from itertools import chain
from keyword import iskeyword
from keyword import kwlist
from numbers import Real
from operator import contains
from typing import ByteString
from typing import Dict
from typing import List
from typing import MutableSequence
from typing import Sequence
from typing import Union

import tatsu
from more_itertools import collapse
from tatsu.exceptions import FailedParse

# grammar symbols
OPERATORS = tuple("+-*/^%")
BRACKETS = tuple("()")
LIST_SEPARATOR = tuple(",")
LETTERS_LOWER = tuple("abcdefghijklmnopqrstuvwxyz")
LETTERS_UPPER = tuple(l.upper() for l in LETTERS_LOWER)
LETTERS_ALL = LETTERS_LOWER + LETTERS_UPPER
DIGITS = tuple("0123465789")
UNDERSCORE = tuple("_")
VARIABLE_NAME = LETTERS_ALL + UNDERSCORE + DIGITS
SPACE = tuple(" ")

ALLOWED_SYMBOLS = (
    LETTERS_ALL
    + DIGITS
    + BRACKETS
    + LIST_SEPARATOR
    + UNDERSCORE
    + SPACE
    + OPERATORS
)

# template includes placeholders for optional items. In the comments are
# examples of what should be in the placeholder.
GRAMMAR_TEMPLATE = r"""
    @@grammar :: openepda_expr
    @@whitespace :: / +/

    start = list $ ;
    list = expr ',' list | expr ;
    expr = expr {{('+' | '-') term}} | term ;
    term = term '%' factor | term '*' factor | term '/' factor | factor ;
    factor = factor "{power_op}" power | power ;  # association right or left
    power = '+' power | '-' power | base ;
    base = number 
           {func2_base}  # | func2 '(' expr ',' expr ')'       
           {func1_base}  # | func1 '(' expr ')'
           {allow_skip_parenth}  # | func1 power  # optional 
           {allow_parenth_after_const}  # | const '(' ')'  # optional 
           {const_base}  # | const  # optional 
           | var
           | "(" list ")" ;
               
    number = /\d*\.\d+[eE][+-]?\d+/ 
             | /\d+\.?[eE][+-]?\d+/ 
             | /\d*\.\d+/ 
             | /\d+\.?/ ;           
 
    {const_list}  # const = 'e' | 'pi' ; 
    
    {func1_list}  # func1 = 'abs' | 'acos' | 'asin' | 'atan' | 'ceil' | 'cos' 
                  # | 'cosh' |'exp' | 'fac' | 'floor' | 'ln' | 'log' | 'log10'  
                  # | 'sin' |'sinh' | 'sqrt' | 'tan' | 'tanh' ;
    
    {func2_list}  # func2 = 'atan2' | 'pow' | 'npr' | 'ncr' ;
    
    # var = /[a-zA-Z][a-zA-Z0-9_]*/ ;
    var = {exclude} var:/[a-zA-Z][a-zA-Z0-9_]*/;
"""

CHOICES_TEMPLATE = "{name} = {choices_str} ;"
EXCLUDE_TEMPLATE = "! ( {choices_str} )"

# Predefined identifiers
CONSTANTS_MAP = {"pi": ("math", "pi"), "e": ("math", "e")}
CONSTANTS = tuple(CONSTANTS_MAP.keys())

# Mapping of functions available in the expressions, 1 and 2 parameters
STANDARD_FCNS_1_ARG_MAP = {
    "abs": ("math", "fabs"),
    "acos": ("math", "acos"),
    "asin": ("math", "asin"),
    "atan": ("math", "atan"),
    "ceil": ("math", "ceil"),
    "cos": ("math", "cos"),
    "cosh": ("math", "cosh"),
    "exp": ("math", "exp"),
    "fac": ("math", "factorial"),
    "floor": ("math", "floor"),
    "ln": ("math", "log"),
    "log": ("math", "log"),  # option to use log10
    "log10": ("math", "log10"),
    "sin": ("math", "sin"),
    "sinh": ("math", "sinh"),
    "sqrt": ("math", "sqrt"),
    "tan": ("math", "tan"),
    "tanh": ("math", "tanh"),
}
STANDARD_FCNS_1_ARG = tuple(STANDARD_FCNS_1_ARG_MAP.keys())

STANDARD_FCNS_2_ARG_MAP = {
    "atan2": ("math", "atan2"),
    "pow": ("math", "pow"),
    "ncr": ("scipy.special", "comb"),
    "npr": ("scipy.special", "perm"),
}
STANDARD_FCNS_2_ARG = tuple(STANDARD_FCNS_2_ARG_MAP.keys())

# Aliases under which the functions can be imported from math
FCN_NAME_ALIASES = {
    "abs": "fabs",
    "fac": "factorial",
    "ln": "log",
    "log": "log10",
}


class OpenEpdaError(Exception):
    pass


class OpenEpdaExpressionError(Exception):
    pass


class _BaseParser(object):
    def __init__(
        self,
        allow_skip_parenth=False,
        use_nat_log=True,
        power_from_right=True,
        use_caret_for_power=True,
        constants=None,
        functions_1=None,
        functions_2=None,
        allow_parenth_after_const=False,
        parameters=(),
        exclude=(),
    ):
        """Create a parser.

        `constants`, `functions_1`, `functions_2` are dictionaries in the
        form: {'expression_name': ('module_name', 'module_attribute')}.
        `parameters` are just an iterable with variable identifiers.
        Their values are submitted to the eval method.

        Parameters
        ----------
        allow_skip_parenth : bool
            if True, in expressions with function of one argument
            (e.g. 'sin 0.5') parentheses can be omitted. If False,
            parentheses are required. Note, that if this option is True,
            parsing of the expression is possible, but evaluation is not
            currently supported.
        use_nat_log : bool
            if True, `log` evaluates to the natural logarithm. Otherwise,
            it is log10.
        power_from_right : bool
            if True, the following association rules are used for
            exponentiation: `a^b^c == a^(b^c)` and `-a^b == -(a^b)`. This is
            default Python behavior. If False, the following rules are used:
            `a^b^c == (a^b)^c` and `-a^b == (-a)^b`. Note, that if this
            option is True, parsing of the expression is possible, but
            evaluation is not currently supported.
        use_caret_for_power : bool
            if True, caret symbol '^' will be used for power operation.
            Otherwise, a standard python operator '**' is used.
        constants : dict
            constants to be allowed in the expression.
        functions_1 : dict
            list of functions of 1 argument to be allowed in the expression.
        functions_2 : dict
            list of functions of 2 arguments to be allowed in the expression.
        parameters : list
            list of parameter identifiers to be allowed in the expression.
        exclude : iterable of str
            identifiers to be excluded from the final list
        """
        # parsing options
        self._allow_parenth_after_const = allow_parenth_after_const
        self._use_caret_for_power = use_caret_for_power

        self._allow_skip_parenth = allow_skip_parenth  # not used
        self._power_from_right = power_from_right  # not used

        # evaluation option
        self._use_nat_log = use_nat_log

        self._exclude = exclude

        constants = constants or {}
        functions_1 = functions_1 or {}
        functions_2 = functions_2 or {}

        # member variables
        self._parser = None
        self._imported_symbols = {}

        # import and add symbols
        self._check_symbols(constants.keys())
        added_symbols = self._add_symbols(constants, exclude=self._exclude)
        self._constants = added_symbols

        self._check_symbols(functions_1.keys())
        added_symbols = self._add_symbols(functions_1, exclude=self._exclude)
        self._f1 = added_symbols

        self._check_symbols(functions_2.keys())
        added_symbols = self._add_symbols(functions_2, exclude=self._exclude)
        self._f2 = added_symbols

        self._check_symbols(parameters)
        self._vars = parameters

        # create parser
        self.compile_parser()

    def _check_symbols(self, symbols):
        """Check if the symbols are allowed in the expression

        The check will not pass if:
            - there are repetitions inside symbols
            - any symbol was already defined before in this parser
            - any symbol is the same as a python keyword

        Parameters
        ----------
        symbols : iterable of str

        Returns
        -------
        True

        Raises
        ------
        ValueError:
            if the check fails.
        """
        if len(symbols) == 0:
            return True

        # check there is no repeated symbols
        repeated_symb, no_repetitions = Counter(symbols).most_common()[0]
        if no_repetitions > 1:
            raise OpenEpdaExpressionError(
                'Symbol "{}" is defined more than once ({} times).'.format(
                    repeated_symb, no_repetitions
                )
            )

        # Check each symbol
        for s in symbols:
            if iskeyword(s):
                # python keywords
                raise OpenEpdaExpressionError(
                    'Symbol "{}" is a reserved keyword in python.'.format(s)
                )
            elif s in self._imported_symbols.keys():
                # already defined symbols
                raise OpenEpdaExpressionError(
                    'Symbol "{}" is already defined.'.format(s)
                )
        return True

    @property
    def reserved_identifiers(self):
        return tuple(self._imported_symbols)

    def _add_symbols(self, symbols, exclude=()):
        """

        Parameters
        ----------
        symbols : dict

        exclude : iterable of str

        Returns
        -------
        tuple of str
            list of imported symbols
        """
        res = make_symbol_mapping(symbols, exclude=exclude)
        self._imported_symbols.update(res)
        return tuple(res.keys())

    def compile_parser(self):
        """Create a parser.

        This is an expensive action, so preferably has to be done only
        once. Parsing itself is much faster.
        """
        # Create data to fill grammar template
        # lists of symbols
        const_list = make_grammar_choice_item("const", self._constants)
        func1_list = make_grammar_choice_item("func1", self._f1)
        func2_list = make_grammar_choice_item("func2", self._f2)

        const_base = r"| const" if const_list else ""
        func1_base = r"| func1 '(' expr ')'" if func1_list else ""
        func2_base = r"| func2 '(' expr ',' expr ')'" if func2_list else ""

        exclude = make_grammar_exclude_item(
            kwlist, self._constants, self._f1, self._f2
        )

        # optional items
        if self._allow_skip_parenth and func1_list:
            allow_skip_parenth = "| func1 power"
        else:
            allow_skip_parenth = ""

        if self._allow_parenth_after_const and const_list:
            allow_parenth_after_const = "| const '(' ')'"
        else:
            allow_parenth_after_const = ""

        power_op = "^" if self._use_caret_for_power else "**"

        grammar = GRAMMAR_TEMPLATE.format(
            power_op=power_op,
            const_base=const_base,
            const_list=const_list,
            func1_base=func1_base,
            func1_list=func1_list,
            func2_base=func2_base,
            func2_list=func2_list,
            allow_skip_parenth=allow_skip_parenth,
            allow_parenth_after_const=allow_parenth_after_const,
            exclude=exclude,
        )

        # print(grammar)
        self._parser = tatsu.compile(grammar)

    def parse(self, expr):
        """Parse the expression.

        Parameters
        ----------
        expr : str
            expression to be parsed

        Returns
        -------
        tatsu.contexts.closure
            AST for the expression
        """
        if not self._parser:
            raise OpenEpdaExpressionError("Parser has not been created yet.")

        if expr == "":
            raise OpenEpdaExpressionError("Cannot parse empty expression")

        try:
            ast = self._parser.parse(expr)
        except FailedParse as e:
            raise OpenEpdaExpressionError(
                "error parsing expression '{}': {}.".format(expr, e)
            )
        return ast

    def get_parameter_names(self, expr):
        """Parse expression and get the names of parameters used in it.

        Parameters
        ----------
        expr : str
            expression to the parsed

        Returns
        -------
        list of str
            list of parameter names present in the expression
        """
        ast = self.parse(expr)
        flat_ast = collapse(ast, base_type=dict)
        param_names = [
            item["var"] for item in flat_ast if isinstance(item, dict)
        ]
        return param_names

    def eval(self, expr, params=None):
        """

        Parameters
        ----------
        expr : str
            expression to be evaluated
        params : Dict[str, Any]
            {'parameter_name': parameter_value}

        Returns
        -------
        Any
            evaluation result
        """
        if not self._power_from_right:
            raise NotImplementedError(
                "Evaluations of expressions with left association of "
                "exponentiation is not implemented."
            )

        if self._allow_skip_parenth:
            raise NotImplementedError(
                "Evaluations of expressions with missing parentheses "
                "for a function of one argument is not implemented."
            )

        params = params or {}
        self.check_missing_params(expr, params.keys())

        if self._allow_parenth_after_const:
            for c in self._constants:
                expr = re.sub(r"{}\( *\)".format(c), "{}".format(c), expr)

        if self._use_caret_for_power:
            expr = expr.replace("^", "**")

        res = eval(expr, self._imported_symbols, params)
        return res

    def check_missing_params(self, expr, params):
        """Verify that all params that are present in the expr are present.

        Parameters
        ----------
        expr : str

        params : iterable of str
            allowed parameters

        """
        actual_param_names = self.get_parameter_names(expr)
        missing_params = set(actual_param_names) - set(params)
        if missing_params:
            raise OpenEpdaExpressionError(
                "expression '{}' contains undefined params: {}.".format(
                    expr, missing_params
                )
            )
        else:
            return True


def make_choice_list(*choices, exclude=()):
    """Combine choices into a single list with unique items.

    Some of the choices can be excluded.

    Parameters
    ----------
    choices : iterable of str
        each iterable contains a list of functions to be included into
        the symbols.
    exclude: iterable of str
        list of symbols to be excluded.
    Returns
    -------
    list of str
        sorted list of symbols, where all choices are united, and
        strings from exclude parameter are removed.
    """
    choices_clean = sorted(set(chain.from_iterable(choices)))

    ch = list(c for c in choices_clean if c not in exclude)
    return ch


def make_symbol_mapping(mapping, exclude=()):
    """Import all symbols from the mapping using their name

    Parameters
    ----------
    mapping : Dict[str, Tuple(str, str)]
        keys are symbol names, values are 2-tuples. value[0] is a
        module name, value[1] is attribute name to be imported from this
        module and to be used under symbol name.
    exclude : iterable of str
        identifiers to be excluded from the final list

    Returns
    -------
    Dict[str, Any]
        keys are imported symbol names, values are imported entities,
        e.g. function, parameters, etc.
    """
    res = {}

    for symbol_name, (module_name, attr) in mapping.items():
        if symbol_name in exclude:
            continue
        m = import_module(module_name)
        try:
            symbol = getattr(m, attr)
        except AttributeError as e:
            raise ValueError(
                "Incorrect data for symbol '{}', cannot import it: {}.".format(
                    symbol_name, e
                )
            )
        res.update({symbol_name: symbol})
    return res


def make_choice_str(choices=None):
    """

    Parameters
    ----------
    choices : iterable of str

    Returns
    -------
    str
    """
    if choices:
        choices_str = "' | '".join(choices)
        if choices_str != "":
            choices_str = "'{}'".format(choices_str)
    else:
        choices_str = ""

    return choices_str


def make_grammar_choice_item(name="", choices=None):
    choices_str = make_choice_str(choices)

    if choices_str != "":
        res = CHOICES_TEMPLATE.format(name=name, choices_str=choices_str)
    else:
        res = ""
    return res


def make_grammar_exclude_item(*choices):
    choices_flat = chain.from_iterable(choices)
    choices_str = make_choice_str(choices_flat)

    if choices_str != "":
        res = EXCLUDE_TEMPLATE.format(choices_str=choices_str)
    else:
        res = ""
    return res


def check_parameter_name(name, reserved=()):
    """Check if name can be used as identifier in the expression.

    - Contains letters, underscores, or digits
    - Starts with a letter
    - Is not a python keyword
    - Is not one of the reserved identifiers

    Parameters
    ----------
    name : str
        identifier to be checked
    reserved : iterable of str
        identifiers which are reserved and are therefore illegal,
        e.g. for other functions.

    Returns
    -------
    bool
        True if name is a correct identifier
    """
    if not (name[0] in LETTERS_LOWER or name[0] in LETTERS_UPPER):
        # First symbol
        raise OpenEpdaExpressionError(
            'First symbol of the parameter name "{}" is not one of allowed'
            " symbols.".format(name)
        )
    elif set(name) - set(VARIABLE_NAME):
        # Other symbols
        raise OpenEpdaExpressionError(
            'Parameter name "{}" contains symbol(s) which is (are) not '
            "letters or an underscore.".format(name)
        )
    elif iskeyword(name):
        # python keywords
        raise OpenEpdaExpressionError(
            'Parameter name "{}" is a reserved keyword in python.'.format(name)
        )
    elif name in reserved:
        raise OpenEpdaExpressionError(
            'Parameter name "{}" is a reserved identifier.'.format(name)
        )
    else:
        return True


def check_expr(expr):
    if not isinstance(expr, str):
        raise TypeError(
            "Expression is not a str type. It is {}.".format(type(expr))
        )

    expr = expr.replace(" ", "")  # remove whitespaces

    if any(map(partial(contains, expr), kwlist)):
        raise OpenEpdaExpressionError(
            "Expression contains a reserved python keyword."
        )
    elif not all(map(partial(contains, ALLOWED_SYMBOLS), expr)):
        raise OpenEpdaExpressionError("Expression contains illegal symbols.")
    else:
        return True


openepda_parser = _BaseParser(
    functions_1=STANDARD_FCNS_1_ARG_MAP,
    functions_2=STANDARD_FCNS_2_ARG_MAP,
    constants=CONSTANTS_MAP,
    exclude="ln",
)

TNestedSequence = Union[str, Real, Sequence["TNestedSequence"]]
TNestedRealSequence = Union[Real, Sequence["TNestedRealSequence"]]
TNestedRealList = Union[Real, List["TNestedRealList"]]


def evaluate_expression(
    expr: TNestedSequence, params=None
) -> Union[Real, TNestedRealList]:
    """Evaluate expression if necessary and return a numeric value.

    Parameters
    ----------
    expr : TNestedSequence
        expression to be checked and evaluated if needed. Can be a sequence
        of expressions, or a sequence of sequences...
    params : Optional[Dict[str, Any]]
        Parameter values to be used during expression evaluation.
        We do not restrict parameter values to the Real type, however,
        this could make sense. It depends on whether the functions used in the
        expression are able to produce numeric output for non-numeric inputs.

    Returns
    -------
    res : Union[Real, TNestedRealList]
        the input expression if it is a number, or the expression evaluation
         result if it is a number.

    Raises
    ------
    TypeError
        In cases when the expression cannot be evaluated, or when it returned
        a non-numeric result.
    """
    if isinstance(expr, Real):
        res = expr
    elif isinstance(expr, str):
        res = openepda_parser.eval(expr, params=params)
        if not isinstance(res, Real):
            raise TypeError(
                "Expression '{}' was evaluated, and a non-numeric type {} "
                "was returned. Evaluation result: {}.".format(
                    expr, type(res), res
                )
            )
    elif isinstance(expr, (MutableSequence, list)):
        res = [evaluate_expression(e, params=params) for e in expr]
    elif isinstance(expr, tuple):
        res = tuple(evaluate_expression(e, params=params) for e in expr)
    elif isinstance(expr, Sequence) and not isinstance(ByteString):
        res = tuple(evaluate_expression(e, params=params) for e in expr)
    else:
        raise TypeError(
            "Provided input '{}' has type {}. It is not a string or a number, "
            "and therefore cannot be evaluated as a numeric value. ".format(
                expr, type(expr)
            )
        )
    return res



# -*- coding: utf-8 -*-
"""

Author: Dima Pustakhod
Copyright: 2020, TU/e - PITC and authors

Changelog
---------
2020.04.28
    Initial commit
"""

try:
    from ruamel.yaml import YAML

    yaml = YAML(typ="safe")
    yaml.default_flow_style = False
    safe_load = yaml.load
    dump_long = yaml.dump

    yaml2 = YAML(typ="safe")
    dump_short = yaml2.dump
except ImportError:
    from yaml import safe_load
    from yaml import dump
    from functools import partial

    dump_long = partial(dump, default_flow_style=False)
    dump_short = partial(dump, default_flow_style=True)


REQUIRED_NAZCA_VERSION = "0.5.11"

try:
    import nazca as nd

    if nd.__version__ != REQUIRED_NAZCA_VERSION:
        print(
            "This version of openepda requires nazca-{}. "
            "Version {} is installed.".format(
                REQUIRED_NAZCA_VERSION, nd.__version__
            )
        )
        HAS_NAZCA_DESIGN = False
    else:
        nd.clear_layout()
        # nd.clear_xsections()
        # nd.clear_layers()
        HAS_NAZCA_DESIGN = True
except ImportError:
    HAS_NAZCA_DESIGN = False


REQUIRED_KLAYOUT_VERSION = "0.26.4"

try:
    import klayout.dbcore as dbcore
    HAS_KLAYOUT = True
    Cell = dbcore.Cell
    Layout = dbcore.Layout
    LayerInfo = dbcore.LayerInfo
except ImportError:
    HAS_KLAYOUT = False



# -*- coding: utf-8 -*-
"""openepda.geometry.py

This file contains routines to work with geometrical objects: points and
lines.

Author: Dima Pustakhod
Copyright: 2020, TU/e - PITC and authors

Changelog
---------
2020.04.29
    Initial commit
"""
from numbers import Real
from typing import List
from typing import Sequence
from typing import Tuple
from typing import Union as U

import numpy as np

TPt = Tuple[Real, Real]
TPts = Sequence[TPt]
TPtList = List[TPt]


import numbers
import math


def isclose(a, b, *, rel_tol=None, abs_tol=None) -> bool:
    """Perform safe comparison of a and b

    If a and b are numbers.Number, math.isclose function is used.
    If a and b are sets, they are sorted and compared element-wise recursively.
    If a and b are dicts, keys and corresponding items are compared recursively.
    if a and b are iterables, they are compared recursively element-wise.
    For all other types of a and b, the function uses standard `==` operator.

    Examples
    --------
    >>> isclose(1.0, 1.001, abs_tol=1e-3)
    True
    >>> isclose(1.0, 1.01, abs_tol=1e-3)
    False
    >>> isclose('a', 'a', abs_tol=1e-3)
    True
    >>> isclose('a', 'b', abs_tol=1e-3)
    False
    """
    if isinstance(a, numbers.Real) and isinstance(b, numbers.Real):
        kwargs = {}
        if rel_tol is not None:
            kwargs.update({"rel_tol": rel_tol})
        if abs_tol is not None:
            kwargs.update({"abs_tol": abs_tol})
        res = math.isclose(a, b, **kwargs)
    elif isinstance(a, str) and isinstance(b, str):
        res = bool(a == b)
    elif isinstance(a, set) and isinstance(b, set):
        res = all((isclose(ai, bi) for ai, bi in zip(a, b)))
    elif isinstance(a, dict) and isinstance(b, dict):
        try:
            res = bool(len(a) == len(b))
            if res:
                res = isclose(sorted(set(a.keys())), sorted(set(b.keys())))
            if res:
                res = all((isclose(a[k], b[k]) for k in a.keys()))
        except Exception:
            res = False
    else:
        try:
            res = bool(len(a) == len(b))
            if res:
                res = all((isclose(ai, bi) for ai, bi in zip(a, b)))
        except Exception:
            res = bool(a == b)
    return res


def _are_points_equal(p1: TPt, p2: TPt) -> bool:
    """Check if the points are close to each other. Not type safe!
    """
    return math.isclose(p1[0], p2[0]) and math.isclose(p1[1], p2[1])


def _is_first_point_smaller(p1: TPt, p2: TPt) -> True:
    """ Check if first point is smaller than the second

    Smaller means located a) more to the left, and b) more to the bottom.
    b) is checked only if a) is inconclusive.

    Parameters
    ----------
    p1: tuple
        contains two coordinates of the point
    p2: tuple
        contains two coordinates of the point
    """
    if math.isclose(p1[0], p2[0]):
        return p1[1] < p2[1]
    else:
        return p1[0] < p2[0]


def count_edges(points: TPts) -> int:
    """Count number of edges in the line.

    Returns
    -------
    int
        -1 if no points are given, 0 if a single point is given.
    """
    return len(points) - 1


def count_vertices(points: TPts) -> int:
    """Count points in the line
    """
    return len(points)


def ensure_line_valid(points: TPts, min_point_count=2) -> TPtList:
    """ Return a point list representing a valid line.

    Parameters
    ----------
    points : list of tuples
        each tuple contains two coordinates of a point
    min_point_count : int
        1 or 2. number of points

    """
    if min_point_count not in (1, 2):
        raise ValueError(
            "min_point_count must have a value of 1 or 2."
            "{} is given".format(min_point_count)
        )

    n_lines = count_edges(points)
    if n_lines < 0:
        return [(0, 0)] * min_point_count
    elif n_lines == 0:
        # single point. return it several times
        return list(points) * min_point_count
    else:
        return list(points)


def is_line_string_closed(points: TPts, raise_error=True) -> bool:
    """ Determine if the line is closed.

    The line given by a sequence of points is closed, if the first and
    the last point are the same. A pen drawing them returns to the same
    location or doesn't move at all.

    Parameters
    ----------
    points : list of points
        a point can be a Point, or a tuple / list of two coordinates
    raise_error : bool
        if True, raise error if points is an empty list. If False, return False.
    """
    n_lines = count_edges(points)
    if n_lines < 0:
        # empty array
        if raise_error:
            raise ValueError("List of points must contain at least one point.")
        else:
            return False
    if n_lines == 0:
        # single point, always closed
        return True
    else:
        return _is_line_string_closed([points[0], points[-1]])


def _is_line_string_closed(points: TPts) -> bool:
    """Determine if the line is closed. Not type safe!

    The line given by a sequence of points is closed, if the first and
    the last point are the same.

    Parameters
    ----------
    points : TPts
        each tuple contains two coordinates of a point
    """
    return _are_points_equal(points[0], points[-1])


def ensure_line_is_closed(
    points: TPts, raise_error=True, min_point_count=2
) -> List[TPt]:
    """ Return a closed line.

    Check if line is closed and add the starting point to the end if it is not.

    Parameters
    ----------
    points : TPts
        each tuple contains two coordinates of a point
    raise_error : bool

    min_point_count : int

    """
    n_lines = count_edges(points)
    if n_lines < 0:
        if raise_error:
            raise ValueError("List of points must contain at least 3 points.")
        else:
            return ensure_line_valid([(0, 0)], min_point_count)
    elif n_lines == 0:
        # single point, always closed. return it several times
        return ensure_line_valid(points, min_point_count)
    else:
        if is_line_string_closed(points):
            return list(points)
        else:
            return list(points) + [points[0]]


def calculate_area(points: TPts) -> U[int, float]:
    """ Calculate the area of simple polygon given by a list of tuples.

    Sign is positive if points are sorted counter clockwise.

    Parameters
    ----------
    points : list of tuples
        each tuple contains two coordinates of a point

    """
    points = np.array(ensure_line_is_closed(points, raise_error=False,))
    area = (
        sum(points[:-1, 0] * points[1:, 1] - points[1:, 0] * points[:-1, 1]) / 2
    )
    return int(area) if area.is_integer() else float(area)


def is_oriented_cw(points: TPts) -> bool:
    """ Calculate the line orientation.

    Parameters
    ----------
    points : list of tuples
        each tuple contains two coordinates of a point

    Return
    ------
    o : bool
        False if line if oriented counter clockwise,
        True if line is oriented clockwise.
    """
    a = calculate_area(points)

    if a < 0:
        return True
    elif a > 0:
        return False
    else:
        return None


def sort_polygon_vertices(points: TPts) -> TPtList:
    """ Sort polygon vertices list.

    Parameters
    ----------
    points : TPts
        each tuple contains two coordinates of a point

    """
    n_pts = count_vertices(points)
    if n_pts <= 1:
        return list(points)
    else:
        return _sort_polygon_vertices(points)


def _sort_polygon_vertices(points: TPts) -> TPtList:
    """Sort polygon vertices list.

    The first point will be the most SW point. The order of other points
    is preserved.

    Parameters
    ----------
    points : list of tuples
        each tuple contains two coordinates of a point

    """
    min_ind, min_p = 0, points[0]
    for i, pt in enumerate(points):
        if _is_first_point_smaller(pt, min_p):
            min_ind, min_p = i, pt

    pts_sorted = points[min_ind:] + points[0:min_ind]
    return pts_sorted


def orient_polygon_vertices(
    points: TPts, clockwise=True, keep_first=False
) -> List[TPt]:
    """Change orientation of the polygon vertices list.

    If orientation is different from required, the point order is inverted.
    No sorting is done.

    Parameters
    ----------
    points : TPts
        each tuple contains two coordinates of a point
    clockwise : bool
        Direction of the vertices orientation.
    keep_first : bool
        Flag to force keeping the first point the same

    """
    o = is_oriented_cw(points)  #

    if o is None or o == clockwise:
        # Orientation not defined or it is correct
        return list(points)
    elif o != clockwise:
        if keep_first:
            return [points[0]] + list(reversed(points[1:]))
        else:
            return list(reversed(points))
    else:
        assert False


def process_polygon_vertices(
    points: TPts,
    sort=True,
    orient=True,
    clockwise=True,
    make_valid=True,
    close=False,
) -> TPtList:
    """
    Parameters
    ----------
    points : TPts

    """
    if make_valid:
        points = ensure_line_valid(points, min_point_count=2)

    if orient:
        points = orient_polygon_vertices(points, clockwise, False)

    if sort:
        points = sort_polygon_vertices(points)

    if close:
        points = ensure_line_is_closed(
            points, raise_error=False, min_point_count=2
        )

    return points



# -*- coding: utf-8 -*-
"""openepda.validators.py

This file contains validators for the data formats defined by
the openEPDA.

Author: Dima Pustakhod
Copyright: 2020, TU/e - PITC and authors
"""
import os
from itertools import chain
from numbers import Real

from .expr import OpenEpdaExpressionError
from .expr import check_parameter_name
from .expr import openepda_parser

try:
    from ruamel.yaml import YAML

    yaml = YAML(typ="safe")
    safe_load = yaml.load
except ImportError:
    from yaml import safe_load


module_fpath = os.path.abspath(os.path.join(__file__, ".."))

CDF_VERSIONS = ["0.1", "0.2", "0.3"]
CDF_SCHEMA_NAME_TEMPLATE = "/schemas/cdf_schema_v{version}.yaml"

UPDK_SBB_VERSIONS = ["0.2"]
UPDK_SBB_SCHEMA_NAME_TEMPLATE = "/schemas/updk_sbb_schema_v{version}.yaml"


def is_cdf_valid(data, version="latest", raise_error=False, full_output=False):
    """Check if the data is a correct CDF of a given version.

    JSON Schema is used for validation. See http://json-schema.org

    Parameters
    ----------
    data : any
        data structure to be validated
    version : str
        CDF format version, 'X.Y' or 'latest'
    raise_error : bool
        If True, a jsonschema.ValidationError is raised in case of wrong
        format. If False, a boolean result will be returned (with
        optional message).
    full_output : bool
        If False, only resulting boolean value is returned. Otherwise,
        additional string with a validation message is returned.

    Returns
    -------
    bool
        validation result
    str
        validation message (optional)

    Examples
    --------
    >>> with open('tests/_test_data/cdf_correct_v0.3.cdf') as s:
    ...     data = safe_load(s)
    >>> is_cdf_valid(data, raise_error=False)
    True

    >>> is_cdf_valid(data, version='0.3', raise_error=False)
    True

    """
    from jsonschema import validate, ValidationError

    allowed_versions = ["latest"] + CDF_VERSIONS
    if version not in allowed_versions:
        raise ValueError(
            "Unknown version of the CDF specified: {version}. "
            "Allowed versions: {allowed_versions}.".format(
                version=version, allowed_versions=allowed_versions
            )
        )
    if version == "latest":
        version = CDF_VERSIONS[-1]

    cdf_schema_file_name = module_fpath + CDF_SCHEMA_NAME_TEMPLATE.format(
        version=version
    )

    with open(cdf_schema_file_name) as s:
        schema = safe_load(s)

    # default reply
    result, msg = True, "Data is valid CDF v.{version}.".format(version=version)

    # Validate against the schema
    if raise_error:
        validate(instance=data, schema=schema)
    else:
        try:
            validate(instance=data, schema=schema)
        except ValidationError as e:
            result = False
            msg = str(e)

    if full_output:
        return result, msg
    else:
        return result


def is_updk_sbb_valid(
    data, version="latest", raise_error=False, full_output=False
):
    """Check if the data structure is a correct uPDK SBB.

    JSON Schema is used for validation. See http://json-schema.org .
    Validation runs against a specification with a given version.

    Parameters
    ----------
    data : Dict
        data structure to be validated
    version : str
        format version, 'X.Y' or 'latest'
    raise_error : bool
        If True, a jsonschema.ValidationError is raised in case of wrong
        format. If False, an boolean result will be returned (with
        optional message).
    full_output : bool
        If False, only resulting boolean value is returned. Otherwise,
        additional string with a validation message is returned.

    Returns
    -------
    bool
        validation result
    str
        validation message (optional)

    Examples
    --------
    >>> with open('tests/_test_data/sbb_correct_v0.2.yaml') as s:
    ...     data = safe_load(s)
    >>> is_updk_sbb_valid(data, raise_error=True)
    True

    """
    from jsonschema import validate, ValidationError

    allowed_versions = ["latest"] + UPDK_SBB_VERSIONS
    if version not in allowed_versions:
        raise ValueError(
            "Unknown version of the uPDK specified: {version}. "
            "Allowed versions: {allowed_versions}.".format(
                version=version, allowed_versions=allowed_versions
            )
        )
    if version == "latest":
        version = UPDK_SBB_VERSIONS[-1]

    updk_schema_file_name = module_fpath + UPDK_SBB_SCHEMA_NAME_TEMPLATE.format(
        version=version
    )

    with open(updk_schema_file_name) as s:
        schema = safe_load(s)

    # default reply
    result, msg = True, "Data is valid uPDK SBB v.{}.".format(version)

    # validate against the schema
    if raise_error:
        validate(instance=data, schema=schema)
    else:
        try:
            validate(instance=data, schema=schema)
        except ValidationError as e:
            result = False
            msg = str(e)

    # validate bb parameter names
    if result:
        if raise_error:
            _check_building_blocks(data["blocks"])
        else:
            try:
                _check_building_blocks(data["blocks"])
            except ValueError as e:
                result = False
                msg = str(e)

    if full_output:
        return result, msg
    else:
        return result


def _check_building_blocks(building_blocks):
    """Check building blocks' params and expressions.

    Parameters
    ----------
    building_blocks : dict

    Returns
    -------
    bool
        True if all BBs are correct

    Raises
    ------
    ValueError
        if check fails
    """
    for bb_name, bb in building_blocks.items():
        _check_building_block(bb_name, bb)


def _check_building_block(bb_name, bb):
    """Check building blocks' params and expressions.

    Currently checking:
        - BB parameter names
        - BB width, length expressions
        - BB bounding box coordinates
        - BB pins width and xya

    Parameters
    ----------
    bb_name : str
        name of the building block.
    bb : dict
        Building block item from the uPDK.

    Returns
    -------
    bool
        True if all BBs are correct

    Raises
    ------
    ValueError
        if check fails
    """
    # Check parameters
    has_params = bool(bb["parameters"])

    if has_params:
        for name, param in bb["parameters"].items():
            _check_parameter(name, param, bb_name)
        bb_params = list(bb["parameters"].keys())
    else:
        bb_params = []

    if "bb_width" in bb:
        _check_bb_num_value(bb["bb_width"], bb_name, params=bb_params)

    if "bb_length" in bb:
        _check_bb_num_value(bb["bb_length"], bb_name, params=bb_params)

    # check bbox
    _check_bbox(bb["bbox"], bb_name, bb_params)

    # check pins
    for pin_name, pin in bb["pins"].items():
        _check_pin(pin_name, pin, bb_name, bb_params)

    for p in ("pin_in", "pin_out"):
        if p in bb:
            if not bb[p] in bb["pins"].keys():
                raise ValueError(
                    "{} '{}' is not present in the list of pins "
                    "for in BB {}.".format(p, bb["pin_in"], bb_name)
                )

    return True


def _check_parameter(name, data, bb_name):
    check_parameter_name(name, reserved=openepda_parser.reserved_identifiers)


def _check_bbox(bbox_data, bb_name, bb_params):
    # check bbox
    for bbox_coord_expr in chain(*bbox_data):
        _check_bb_num_value(bbox_coord_expr, bb_name, params=bb_params)


def _check_bb_num_value(v, bb_name, params=(), parser=openepda_parser):
    """Wrapper around _check_num_value() for BB params.
    """
    try:
        _check_num_value(v, params=params, parser=parser)
    except OpenEpdaExpressionError as e:
        raise ValueError(
            "error in numeric value in BB {}: {}".format(bb_name, e)
        )
    return True


def _check_pin(pin_name, pin_data, bb_name, bb_params):
    _check_pin_num_value(pin_data["width"], pin_name, bb_name, params=bb_params)
    for i in range(3):
        _check_pin_num_value(
            pin_data["xya"][i], pin_name, bb_name, params=bb_params
        )


def _check_pin_num_value(
    v, pin_name, bb_name, params=(), parser=openepda_parser
):
    """Wrapper around _check_num_value() for BB pin params.
    """
    try:
        _check_num_value(v, params=params, parser=parser)
    except OpenEpdaExpressionError as e:
        raise ValueError(
            "error in numeric value in pin {} of BB {}: {}".format(
                pin_name, bb_name, e
            )
        )
    return True


def _check_num_value(v, params=(), parser=openepda_parser):
    """Check if the value is a proper number or expression

    Check is successful if the expression is a number, or if:
        - expression can be parsed using parser
        - it contains only params which are explicitly listed in params

    Parameters
    ----------
    v : num or str
        Expression to be checked.
    params : iterable of str
        list of allowed variable names
    parser : _BaseParser
        parser to be used for the expression parsing

    Returns
    -------
    bool
        True if expr is correct

    Raises
    ------
    ValueError
        if the check is not successful
    """
    if isinstance(v, Real):
        return True
    else:
        return parser.check_missing_params(v, params)


def main():
    import doctest

    os.chdir("{}/..".format(module_fpath))
    doctest.testmod(verbose=False)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
""" test_data_dumper.py

This script tests the OpenEpdaDataDumper class.

Author: Dima Pustakhod

Changelog
---------
2019.07.19
    Initial version

"""
import logging
import numpy as np
import matplotlib.pyplot as plt
from openepda import OpenEpdaDataDumper
from time import time

fname_02 = '_output_data/openepda_data_v.0.2.csv'
fname_03 = '_output_data/openepda_data_v.0.3.csv'

N = 1000000

list3 = ['text, one'] * N
list1 = list(range(N))
array1 = np.linspace(0, 50, num=len(list1))

data = {'string1': 'example1',  # string
        'string2': 'example2',  # string
        'integer1': 1000,  # int
        'integer2': -500,  # int
        'float1': 1.10,  # float
        'float2': 0.1e-18,  # float
        'dict1': {
                'd1s1': 'd1v1', 'd1s2': 'd1v1', 'd1i1': 1000, 'd1i2': -500,
                'd1f1': 1.10, 'd1f2': 0.1e-18},  # dict
        'dict2': {
               'd2s1': 'v1', 'd2s2': 'v1', 'd2i1': 1000, 'd2i2': -500,
               'd2f1': 1.10, 'd2f2': 0.1e-18},  # dict
        'list': list3,
        'l1,3': list1,  # list len1
        'l2': array1,  # np array len1
        # 'l1': 1.10,  # list len2
        # 'l2': 0.1e-18,  # list len2
        }

w = OpenEpdaDataDumper()
# w._l.setLevel(logging.DEBUG)

start = time()
with open(fname_02, 'w') as f:
    w.dump(data, f)

duration = time() - start
print(duration)



import doctest
import unittest
import os

from openepda import validators


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(validators))
    return tests


module_fpath = os.path.abspath(os.path.join(__file__, ".."))


if __name__ == "__main__":
    os.chdir(module_fpath + "/..")
    # print(module_fpath)
    unittest.main()



# -*- coding: utf-8 -*-
"""openepda.black_box.py

This file contains a black box interface and .

Author: Dima Pustakhod
Copyright: 2020, TU/e - PITC and authors
"""
from abc import ABC
from abc import abstractmethod
from numbers import Real
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union as U

from .geometry import process_polygon_vertices, isclose
from .imports import HAS_KLAYOUT
from .imports import safe_load
from .updk import BB
from .updk import DEFAULT_BB_INFO_LAYER
from .updk import DEFAULT_BB_OUTLINE_LAYER
from .updk import DEFAULT_BB_PARAMS_LAYER
from .updk import DEFAULT_BBOX_LAYER
from .updk import DEFAULT_PIN_INFO_LAYER
from .updk import UPDK
from .updk import BBox

if HAS_KLAYOUT:
    from .imports import Cell, Layout

MAPPING = {
    "bbox": (DEFAULT_BBOX_LAYER, 0),
    "bb_outline": (DEFAULT_BB_OUTLINE_LAYER, 0),
    # 'pin_location': (1004, 0),
    # 'pin_stub': (1001, 0),
    "pin_info": (DEFAULT_PIN_INFO_LAYER, 0),
    # 'pin_name': (1002, 0),
    "bb_parameters": (DEFAULT_BB_PARAMS_LAYER, 0),
    "bb_info": (DEFAULT_BB_INFO_LAYER, 0),
}

TParamValue = U[str, bool, int, Real]
TParamValues = Dict[str, TParamValue]
TPinInfoValue = U[str, float, Tuple[float, float, float]]
TPinInfo = Dict[str, TPinInfoValue]


class _BlackBox(ABC):
    """Interface to store black boxes.
    """

    def __init__(self, *args, **kwargs):
        pass

    @property
    @abstractmethod
    def name(self):
        raise NotImplementedError("")

    @property
    @abstractmethod
    def version(self):
        raise NotImplementedError("")

    @property
    def info(self):
        info = {
            "cell_name": self.name,
            "bb_version": self.version,
        }
        return info

    @property
    @abstractmethod
    def bb_outline(self) -> List[Tuple[Real, Real]]:
        raise NotImplementedError("")

    @property
    @abstractmethod
    def bbox(self) -> "BBox":
        raise NotImplementedError("")

    @property
    @abstractmethod
    def pin_names(self) -> Tuple[str]:
        raise NotImplementedError("")

    def get_pin_info(self, pin_name) -> Dict[str, Any]:
        """Returns pin specs

        Returns
        -------
        dict
            Must contain name (str), width (float), xsection_name (str) and
            xya (3-tuple of float)
        """
        pin_infos = self.get_pin_infos()
        assert pin_name in pin_infos
        return pin_infos[pin_name]

    @abstractmethod
    def get_pin_infos(self) -> Dict[str, TPinInfo]:
        raise NotImplementedError("")

    @property
    @abstractmethod
    def parameter_names(self) -> Tuple[str]:
        raise NotImplementedError("")

    def get_parameter_value(self, name) -> TParamValue:
        param_values = self.get_parameter_values()
        assert name in param_values
        return param_values[name]

    @abstractmethod
    def get_parameter_values(self) -> TParamValues:
        raise NotImplementedError("")

    def __eq__(self, other: "_BlackBox") -> bool:
        n = self.name == other.name
        v = self.version == other.version
        outline = isclose(
            self.bb_outline, other.bb_outline, abs_tol=1e-10, rel_tol=1e-10
        )
        bbox = self.bbox == other.bbox

        param_names = set(self.parameter_names) == set(other.parameter_names)
        param_vals = isclose(
            self.get_parameter_values(),
            other.get_parameter_values(),
            abs_tol=1e-10,
            rel_tol=1e-10,
        )

        pin_names = set(self.pin_names) == set(other.pin_names)
        pin_infos = isclose(
            self.get_pin_infos(),
            other.get_pin_infos(),
            abs_tol=1e-10,
            rel_tol=1e-10,
        )

        result = (
            n
            and v
            and outline
            and bbox
            and param_names
            and param_vals
            and pin_names
            and pin_infos
        )
        return result

    @property
    def full_info(self):
        s = (
            f"{self.__class__.__name__}\n"
            f"  Info: {self.info}\n"
            f"  BBox: {self.bbox}\n"
            f"  BB outline: {self.bb_outline}\n"
            f"  Parameters: {self.get_parameter_values()}\n"
            f"  Pins: {self.get_pin_infos()}"
        )
        return s


class uPDKBlackBox(_BlackBox):
    def __init__(self, bb: BB, parameter_values=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bb = bb

        self._bbox = None
        self._bb_outline = None
        self._parameter_values = None
        self._pin_infos = {}

        self.parameter_values = parameter_values

    @property
    def name(self):
        return self._bb.name

    @property
    def version(self):
        return self._bb.version

    @property
    def bb_outline(self) -> List[Tuple[Real, Real]]:
        # set parameters, generate the outline
        if not self._bb_outline:
            self._bb.set_parameters(**self.parameter_values)
            self._bb_outline = process_polygon_vertices(
                self._bb.bb_outline,
                sort=True,
                orient=True,
                clockwise=True,
                make_valid=False,
                close=False,
            )
        return self._bb_outline

    @property
    def bbox(self) -> BBox:
        if not self._bbox:
            self._bb.set_parameters(**self.parameter_values)
            self._bbox = self._bb.bbox
        return self._bbox

    @property
    def pin_names(self) -> Tuple[str]:
        return self._bb.pin_names

    def get_pin_infos(self) -> Dict[str, TPinInfo]:
        if self._pin_infos == {}:
            self._bb.set_parameters(**self.parameter_values)
            for pin_name in self.pin_names:
                pin = self._bb.get_pin(pin_name)
                info = {
                    "name": pin.name,
                    "width": pin.width,
                    "xsection": pin.xsection,
                    "xya": pin.xya,
                }
                self._pin_infos.update({pin_name: info})

        return self._pin_infos

    @property
    def parameter_names(self) -> Tuple[str]:
        return self._bb.parameter_names

    @property
    def parameter_values(self) -> TParamValues:
        return self._parameter_values

    @parameter_values.setter
    def parameter_values(self, values: Optional[Dict[str, TParamValue]]):
        if values is None:
            if self.parameter_names != ():
                raise ValueError(
                    "The parameter values for parameters {} must be "
                    "specified.".format(self.parameter_names)
                )
            self._parameter_values = {}
        else:
            if set(values.keys()) != set(self.parameter_names):
                raise ValueError(
                    "List of provided parameters ({}) does not match the BB "
                    "parameters {}.".format(
                        list(values.keys()), self.parameter_names
                    )
                )
            self._parameter_values = values

    def get_parameter_values(self) -> TParamValues:
        return self.parameter_values


class GdsBlackBox(_BlackBox):
    REQUIRED_MAPPING_KEYS = {
        "bbox",
        "bb_outline",
        "pin_info",
        "bb_parameters",
        "bb_info",
    }

    def __init__(self, cell: "Cell", mapping=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cell = cell
        self._layout: "Layout" = cell.layout()
        self._dbu = self._layout.dbu

        self._name = None
        self._bbox = None
        self._bb_outline = None
        self._parameter_values = None
        self._pin_infos: Dict[str, TPinInfo] = {}

        self._mapping = None
        self._bb_info = None
        self._bb_parameters = None

        mapping = mapping or MAPPING
        self.mapping = mapping
        # self.parameter_values = parameter_values

    def _to_um(self, v):
        return v * self._dbu

    @property
    def mapping(self):
        return self._mapping

    @mapping.setter
    def mapping(self, m):
        missing_keys = set(m.keys()) - self.REQUIRED_MAPPING_KEYS
        if missing_keys:
            raise ValueError(
                f"The layer mapping has the following layers missing: "
                f"{missing_keys}."
            )
        self._mapping = m

    @property
    def bb_parameters(self) -> TParamValues:
        if self._bb_parameters is None:
            i = self._layout.find_layer(*self.mapping["bb_parameters"])
            if i is None:
                raise ValueError(
                    "BB parameters layer is not found in the layout."
                )
            shapes = self._cell.shapes(i)
            n_data = shapes.size()
            if n_data == 0:
                print(f"No parameters found for the BB {self.name}")
                bb_params = {}
            elif n_data > 1:
                raise ValueError(
                    f"More that a single list of parameters is found "
                    f"({n_data})."
                )
            else:
                shape = list(shapes.each())[0]
                if shape.is_text():
                    text = shape.text_string
                else:
                    raise ValueError(
                        "BB parameters layer contains non-text objects."
                    )
                bb_params = safe_load(text)

            self._bb_parameters = bb_params or {}

        return self._bb_parameters

    @property
    def bb_info(self):
        if self._bb_info is None:
            i = self._layout.find_layer(*self.mapping["bb_info"])
            if i is None:
                raise ValueError("BB info layer is not found in the layout.")
            shapes = self._cell.shapes(i)
            n_data = shapes.size()
            if n_data == 0:
                raise ValueError(
                    f"No BB info items are found ({n_data})."
                )
            elif n_data > 1:
                raise ValueError(
                    f"More than a single BB info items are found ({n_data})."
                )
            shape = list(shapes.each())[0]
            if not shape.is_text():
                raise ValueError("BB info layer contains non-text objects.")

            text = shape.text_string
            bb_info = safe_load(text)

            self._bb_info = bb_info

        return self._bb_info

    @property
    def name(self):
        return self.bb_info["cell_name"]

    @property
    def version(self):
        return self.bb_info["bb_version"]

    @property
    def bb_outline(self) -> List[Tuple[Real, Real]]:
        # set parameters, generate the outline
        if not self._bb_outline:
            i = self._layout.find_layer(*self.mapping["bb_outline"])
            if i is None:
                raise ValueError("BB outline layer is not found in the layout.")
            shapes = self._cell.shapes(i)
            n_shapes = shapes.size()
            if n_shapes != 1:
                raise ValueError(
                    f"No. of bboxes must be 1. {n_shapes} is found."
                )
            shape = list(shapes.each())[0]
            if shape.is_box():
                # width = shape.box_width
                # height = shape.box_height
                sw = (self._to_um(shape.box_p1.x), self._to_um(shape.box_p1.y))
                ne = (self._to_um(shape.box_p2.x), self._to_um(shape.box_p2.y))
                pts = [sw, (sw[0], ne[1]), ne, (ne[0], sw[1])]
                # n_pts = len(pts)
            elif shape.is_simple_polygon():
                pts = [
                    (self._to_um(p.x), self._to_um(p.y))
                    for p in shape.simple_polygon.each_point()
                ]
                pts = process_polygon_vertices(
                    pts,
                    sort=True,
                    orient=True,
                    clockwise=True,
                    make_valid=False,
                    close=False,
                )
                # width = shape.bbox().width()
                # height = shape.bbox().height()
                # sw = (shape.bbox().p1.x, shape.bbox().p1.y)
                # ne = (shape.bbox().p2.x, shape.bbox().p2.y)
                # n_pts = len(pts)
            else:
                raise ValueError("BBox is not a box or a simple poly.")

            self._bb_outline = pts

        return self._bb_outline

    @property
    def bbox(self) -> BBox:
        if not self._bbox:
            i = self._layout.find_layer(*self.mapping["bbox"])
            if i is None:
                raise ValueError("Bbox layer is not found in the layout.")
            shapes = self._cell.shapes(i)
            n_shapes = shapes.size()
            if n_shapes != 1:
                raise ValueError(
                    f"No. of bboxes must be 1. {n_shapes} is found."
                )
            shape = list(shapes.each())[0]
            if shape.is_box():
                # width = shape.box_width
                # height = shape.box_height
                sw = (self._to_um(shape.box_p1.x), self._to_um(shape.box_p1.y))
                ne = (self._to_um(shape.box_p2.x), self._to_um(shape.box_p2.y))
                self._bbox = BBox(sw, ne)
            elif shape.is_simple_polygon():
                raise NotImplementedError(
                    "BBox as a simple poly is not implemented yet."
                )
            else:
                raise ValueError("BBox is not a box or a simple poly.")
        return self._bbox

    @property
    def pin_names(self) -> Tuple[str]:
        return tuple(sorted(self.get_pin_infos().keys()))

    def get_pin_infos(self) -> Dict[str, TPinInfo]:
        if self._pin_infos == {}:
            i = self._layout.find_layer(*self.mapping["pin_info"])
            if i is None:
                raise ValueError("Pin info layer is not found in the layout.")
            shapes = self._cell.shapes(i)
            n_data = shapes.size()
            if n_data == 0:
                raise ValueError(f"No pins are found in the cell.")

            for shape in shapes.each():
                if not shape.is_text():
                    raise ValueError(
                        "Pin info layer contains non-text objects."
                    )
                v = shape.text_trans.disp
                xt, yt = self._to_um(v.x), self._to_um(v.y)
                text = shape.text_string
                pin_info = safe_load(text)

                if "name" not in pin_info:
                    raise ValueError(
                        f"The pin info {pin_info} does not contain pin name."
                    )

                pin_name = pin_info["name"]

                if "xya" in pin_info:
                    try:
                        x, y, a = pin_info["xya"]
                    except Exception:
                        raise ValueError(
                            "The pin xya for pin {} must contain a list with "
                            "three numeric values. {} is given.".format(
                                pin_name, pin_info["xya"]
                            )
                        )
                    if not isclose(
                        xt, x, abs_tol=1e-10, rel_tol=1e-10
                    ) or not isclose(yt, y, abs_tol=1e-10, rel_tol=1e-10):
                        raise ValueError(
                            "The pin xya {} for pin {} does not match the "
                            "actual text position ({}, {}).".format(
                                pin_info["xya"], pin_name, xt, yt
                            )
                        )
                    pin_info["xya"] = tuple(pin_info["xya"])

                self._pin_infos.update({pin_name: pin_info})

        return self._pin_infos

    @property
    def parameter_names(self) -> Tuple[str]:
        return tuple(sorted(self.bb_parameters.keys()))

    def get_parameter_values(self) -> TParamValues:
        return self.bb_parameters


class GdsFile(object):
    REQUIRED_MAPPING_KEYS = GdsBlackBox.REQUIRED_MAPPING_KEYS

    def __init__(self, file_name: str, mapping=None):
        if not HAS_KLAYOUT:
            raise RuntimeWarning("The GdsFile class requires klayout package.")

        self._fn = None
        self.file_name = file_name
        self._layout = Layout()
        self._layer_map = self._layout.read(file_name)

        self._top_cell_names = None
        self._cell_names = None
        self._mapping = None

        mapping = mapping or MAPPING
        self.mapping = mapping
        # {cell_name: (Cell, GdsBlackBox)}: For each cell_name, it contains
        # a klayout cell, and a GdsBlackBox. GdsBlackBox are not created
        # before they are accessed by get_gds_black_box() method.
        self._black_boxes: Dict[str, Tuple["Cell", Optional[GdsBlackBox]]] = {
            n: (self._layout.cell(n), None) for n in self.cell_names
        }

    @property
    def mapping(self):
        return self._mapping

    @mapping.setter
    def mapping(self, m):
        missing_keys = self.REQUIRED_MAPPING_KEYS - set(m.keys())
        if missing_keys:
            raise ValueError(
                f"The layer mapping has the following layers missing: "
                f"{missing_keys}."
            )
        self._mapping = m

    @property
    def top_cell_names(self) -> Tuple[str]:
        if self._top_cell_names is None:
            self._top_cell_names = tuple(
                sorted(str(c.name) for c in self._layout.top_cells())
            )
        return self._top_cell_names

    @property
    def cell_names(self) -> Tuple[str]:
        if self._cell_names is None:
            self._cell_names = tuple(
                sorted(str(c.name) for c in self._layout.each_cell())
            )
        return self._cell_names

    def get_cell(self, name) -> "Cell":
        if name not in self.cell_names:
            raise ValueError(f"Cell '{name}' is not found in the GDS file.")

        return self._black_boxes[name][0]

    def get_gds_black_box(self, name: str) -> GdsBlackBox:
        """

        Parameters
        ----------
        name : str
            cell name

        Returns
        -------
        GdsBlackBox
        """
        if name not in self.cell_names:
            raise ValueError(f"Cell '{name}' is not found in the GDS file.")

        if self._black_boxes[name][1]:
            black_box = self._black_boxes[name][1]
        else:
            cell = self.get_cell(name)
            black_box = GdsBlackBox(cell, self.mapping)
            self._black_boxes.update({name: (cell, black_box)})

        return black_box


class BbValidator(object):
    def __init__(self, updk: UPDK, gds_file: GdsFile):
        self._updk = updk
        self._gds_file = gds_file

        self._name_map = None
        self._matched_cell_names = None
        self._updk_black_box_map = None

        assert isinstance(self.matched_cell_names, list)
        assert isinstance(self.unmatched_cell_names, list)

    @staticmethod
    def from_files(updk_name: str, gds_name: str, mapping=None):
        updk = UPDK.from_file(updk_name)
        gds_file = GdsFile(gds_name, mapping=mapping)

        return BbValidator(updk, gds_file)

    @property
    def gds_cell_names(self):
        return self._gds_file.cell_names

    def _find_bb_name_by_cell_name(
        self, cell_name
    ) -> Tuple[List[str], int, str]:
        black_box = self._gds_file.get_gds_black_box(cell_name)
        try:
            black_box_name = black_box.name
        except ValueError as e:
            msg = (
                f"Error getting black box name for cell '{cell_name}':" 
                f" {e}"
            )
            return [], 0, msg

        matching_names = self._find_bb_name_for_black_box_name(black_box_name)

        n = len(matching_names)
        if n == 0:
            msg = (
                f"No matching uPDK block name found for cell '{cell_name}', "
                f"GDS black box name: '{black_box_name}'."
            )
        elif n > 1:
            msg = (
                f"Multiple matching uPDK block names found for cell name "
                f"'{black_box_name}', GDS black box name '{black_box_name}': "
                f"{matching_names}."
            )
        else:
            msg = "Found single matching BB name."

        return matching_names, n, msg

    def _find_bb_name_for_black_box_name(self, black_box_name) -> List[str]:
        bb_names = self._updk.building_block_names

        matching_names = []
        for bb_name in bb_names:
            if self._check_bb_name_match(black_box_name, bb_name):
                matching_names.append(bb_name)

        return matching_names

    @staticmethod
    def _check_bb_name_match(gds_box_name, updk_block_name) -> bool:
        if updk_block_name == gds_box_name:
            return True
        # elif bb_name.startswith(black_box_name):
        #     if not bb_name[len(black_box_name)].isalnum():
        #         matching_names.append(bb_name)

    @property
    def cell_name_map(self) -> Dict[str, Tuple[List[str], int, str]]:
        if self._name_map is None:
            name_map = {}
            for cell_name in self._gds_file.cell_names:
                matching_names, n, msg = self._find_bb_name_by_cell_name(
                    cell_name
                )
                name_map.update({cell_name: (matching_names, n, msg)})
            self._name_map = name_map

        return self._name_map

    @property
    def matched_cell_names(self) -> List[str]:
        if self._matched_cell_names is None:
            matched = []
            for cell_name, match in self.cell_name_map.items():
                if match[1] == 1:
                    matched.append(cell_name)
            self._matched_cell_names = matched

        return self._matched_cell_names

    @property
    def matched_name_map(self) -> Dict[str, str]:
        matched_map = {}
        for cell_name in self.matched_cell_names:
            matched_map.update({cell_name: self.cell_name_map[cell_name][0][0]})
        return matched_map

    @property
    def unmatched_cell_names(self) -> List[str]:
        unmatched = []
        for cell_name, match in self.cell_name_map.items():
            if match[1] != 1:
                unmatched.append(cell_name)
        return unmatched

    def get_unmatched_cells_report(
        self,
    ) -> Tuple[bool, Dict[str, Dict[str, U[int, str, List]]]]:
        report = {}
        passed = True
        for cell_name in self.unmatched_cell_names:
            try:
                names, n, msg = self.cell_name_map[cell_name]
                report.update(
                    {
                        cell_name: {
                            "n_matches": n,
                            "matching_blocks": names,
                            "message": msg,
                        }
                    }
                )
            except Exception as e:
                msg = f"Error fetching match report for cell '{cell_name}': {e}"
                report.update(
                    {
                        cell_name: {
                            "n_matches": None,
                            "matching_blocks": [],
                            "message": msg,
                        }
                    }
                )
                passed = passed and False
        return passed, report

    def get_matching_block_name(self, cell_name) -> str:
        if cell_name not in self.matched_cell_names:
            raise ValueError(
                f"Cell '{cell_name}' does not have a matching uPDK block. Use "
                f"BbValidator.get_unmatched_cells_report() for more details."
            )
        return self.matched_name_map[cell_name]

    def get_gds_black_box(self, cell_name) -> GdsBlackBox:
        return self._gds_file.get_gds_black_box(cell_name)

    def get_parameter_values(self, cell_name) -> Dict:
        black_box = self.get_gds_black_box(cell_name)
        p_values = black_box.get_parameter_values()
        return p_values

    def get_matching_updk_block(self, cell_name) -> BB:
        if cell_name not in self.gds_cell_names:
            raise ValueError(
                f"Cell '{cell_name}' is not present in the GDS file."
            )
        block_name = self.get_matching_block_name(cell_name)
        updk_block = self._updk.get_building_block(block_name)
        return updk_block

    def get_matching_updk_black_box(self, cell_name) -> uPDKBlackBox:
        updk_block = self.get_matching_updk_block(cell_name)
        param_values = self.get_parameter_values(cell_name)
        updk_black_box = uPDKBlackBox(updk_block, param_values)
        return updk_black_box

    def get_matched_cells_report(self) -> Tuple[bool, Dict]:
        report = {}
        passed = True
        for cell_name in self.matched_cell_names:
            try:
                black_gds = self.get_gds_black_box(cell_name)
            except Exception as e:
                msg = (
                    f"Error while obtaining GDS black box for cell "
                    f"'{cell_name}': {e}"
                )
                report.update({cell_name: {"matches": False, "message": msg}})
                passed = passed and False
                continue

            try:
                black_updk = self.get_matching_updk_black_box(cell_name)
            except Exception as e:
                msg = (
                    f"Error while obtaining matching uPDK block for cell "
                    f"'{cell_name}': {e}"
                )
                report.update({cell_name: {"matches": False, "message": msg}})
                passed = passed and False
                continue

            try:
                equal = bool(black_updk == black_gds)

            except Exception as e:
                msg = (
                    f"Error comparing GDS cell '{cell_name}' with the uPDK "
                    f"block '{black_updk.name}': {e}"
                )
                report.update({cell_name: {"matches": False, "message": msg}})
                continue

            params = black_gds.get_parameter_values()
            if equal:
                msg = (
                    f"GDS cell '{cell_name}' matches the uPDK block "
                    f"'{black_updk.name}' with parameters {params}."
                )
                report.update({cell_name: {"matches": True, "message": msg}})
            else:
                msg = (
                    f"GDS cell '{cell_name}' does not match the uPDK block "
                    f"'{black_updk.name}' with parameters {params}."
                )
                report.update(
                    {
                        cell_name: {
                            "matches": False,
                            "message": msg,
                            "gds_block_info": black_gds.full_info,
                            "updk_block_info": black_updk.full_info,
                        }
                    }
                )
                passed = passed and False
        return passed, report



import io
import setuptools

from _version import __version__

import os

if not os.path.exists("README"):
    os.link("README.rst", "README")

with io.open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

long_description = long_description.replace(
    ".. _changes: CHANGES",
    ".. _changes: https://openepda.org/pypi_package/history_of_changes.html",
)
long_description = long_description.replace(
    ".. _authors: AUTHORS", ".. _authors: https://openepda.org/about.html"
)

# Remove badges not available from outside gitlab
long_description = long_description.replace("|ci| |coverage| ", "")

with open("requirements.txt", "r") as fh:
    # remove newline characters
    install_requires = [p[:-1] for p in fh.readlines()]

setuptools.setup(
    name="openepda",
    version=__version__,
    author="Dzmitry Pustakhod",
    author_email="d.pustakhod@tue.nl",
    description="Implementation of open standards for electronic-photonic design automation",
    keywords="EPDA data science electronics photonics design automation standard CAD PDA PIC",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="http://openepda.org",
    packages=setuptools.find_packages(),
    package_data={"openepda": ["schemas/*.*"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        # "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Hardware",
    ],
    install_requires=install_requires,
)
